@@ -14,14 +14,28 @@
 # limitations under the License.
 #
 
+import os
+
+BLE_TRANSPORT_SUPPORTED = False
+COAP_TRANSPORT_SUPPORTED = False
+IP_TRANSPORT_SUPPORTED = False
+
 try:
-    __import__("aioble")
-    BLE_TRANSPORT_SUPPORTED = True
+    if "AIOHOMEKIT_TRANSPORT_BLE" in os.environ:
+        __import__("bleak")
+        BLE_TRANSPORT_SUPPORTED = True
 except ModuleNotFoundError:
-    BLE_TRANSPORT_SUPPORTED = False
+    pass
+
+try:
+    if "AIOHOMEKIT_TRANSPORT_COAP" in os.environ:
+        __import__("aiocoap")
+        COAP_TRANSPORT_SUPPORTED = True
+except ModuleNotFoundError:
+    pass
 
 try:
     __import__("zeroconf")
     IP_TRANSPORT_SUPPORTED = True
 except ModuleNotFoundError:
-    IP_TRANSPORT_SUPPORTED = False
+    pass
