@@ -24,9 +24,9 @@ from cryptography.hazmat.primitives import serialization
 from cryptography.hazmat.primitives.asymmetric import ed25519, x25519
 
 from aiohomekit.crypto import (
+    ChaCha20Poly1305Decryptor,
+    ChaCha20Poly1305Encryptor,
     SrpClient,
-    chacha20_aead_decrypt,
-    chacha20_aead_encrypt,
     hkdf_derive,
 )
 from aiohomekit.exceptions import (
@@ -140,9 +140,8 @@ def perform_pair_setup_part1(
 def validate_mfi(session_key, response_tlv):
     # If pairing method is PairSetupWithAuth there should be an EncryptedData TLV in M4
     # It should have a signature and a certificate from an Apple secure co-processor.
-    decrypted = chacha20_aead_decrypt(
-        bytes(),
-        session_key,
+    decrypted = ChaCha20Poly1305Decryptor(session_key).decrypt(
+        b"",
         b"PS-Msg04",
         bytes([0, 0, 0, 0]),
         response_tlv[TLV.kTLVType_EncryptedData],
@@ -242,14 +241,14 @@ def perform_pair_setup_part2(
     #   Pair-Setup-Encrypt-Info instead of Pair-Setup-Controller-Sign-Info
     ios_device_x = hkdf_derive(
         SrpClient.to_byte_array(session_key),
-        "Pair-Setup-Controller-Sign-Salt",
-        "Pair-Setup-Controller-Sign-Info",
+        b"Pair-Setup-Controller-Sign-Salt",
+        b"Pair-Setup-Controller-Sign-Info",
     )
 
     session_key = hkdf_derive(
         SrpClient.to_byte_array(session_key),
-        "Pair-Setup-Encrypt-Salt",
-        "Pair-Setup-Encrypt-Info",
+        b"Pair-Setup-Encrypt-Salt",
+        b"Pair-Setup-Encrypt-Info",
     )
 
     # if TLV.kTLVType_EncryptedData in response_tlv:
@@ -270,8 +269,8 @@ def perform_pair_setup_part2(
     # taking tge iOSDeviceX as key was reversed from
     # https://github.com/KhaosT/HAP-NodeJS/blob/2ea9d761d9bd7593dd1949fec621ab085af5e567/lib/HAPServer.js
     # function handlePairStepFive calling encryption.encryptAndSeal
-    encrypted_data_with_auth_tag = chacha20_aead_encrypt(
-        bytes(), session_key, b"PS-Msg05", bytes([0, 0, 0, 0]), sub_tlv_b
+    encrypted_data_with_auth_tag = ChaCha20Poly1305Encryptor(session_key).encrypt(
+        b"", b"PS-Msg05", bytes([0, 0, 0, 0]), sub_tlv_b
     )
 
     response_tlv = [
@@ -295,9 +294,8 @@ def perform_pair_setup_part2(
     if TLV.kTLVType_EncryptedData not in response_tlv:
         raise InvalidError("M6: Encrypted data not sent be accessory")
 
-    decrypted_data = chacha20_aead_decrypt(
-        bytes(),
-        session_key,
+    decrypted_data = ChaCha20Poly1305Decryptor(session_key).decrypt(
+        b"",
         b"PS-Msg06",
         bytes([0, 0, 0, 0]),
         response_tlv[TLV.kTLVType_EncryptedData],
@@ -323,8 +321,8 @@ def perform_pair_setup_part2(
 
     accessory_x = hkdf_derive(
         SrpClient.to_byte_array(srp_client.get_session_key()),
-        "Pair-Setup-Accessory-Sign-Salt",
-        "Pair-Setup-Accessory-Sign-Info",
+        b"Pair-Setup-Accessory-Sign-Salt",
+        b"Pair-Setup-Accessory-Sign-Info",
     )
 
     accessory_info = accessory_x + accessory_pairing_id + accessory_ltpk
@@ -350,8 +348,91 @@ def perform_pair_setup_part2(
     }
 
 
+def resume_m1(
+    session_id: bytes, pub_key: bytes, derive: Callable[[bytes, bytes], bytes]
+) -> list[tuple[int, bytes]]:
+    request_key = derive(
+        pub_key + session_id,
+        b"Pair-Resume-Request-Info",
+    )
+
+    logger.debug("resume request key: %s", request_key)
+
+    key = ChaCha20Poly1305Encryptor(request_key)
+
+    auth_tag = key.encrypt(
+        b"",
+        b"PR-Msg01",
+        bytes([0, 0, 0, 0]),
+        b"",
+    )
+
+    logger.debug("auth tag: %s", auth_tag)
+
+    return [
+        (TLV.kTLVType_State, TLV.M1),
+        (TLV.kTLVType_Method, TLV.kTLVMethod_Resume.to_bytes(1, "little")),
+        (TLV.kTLVType_PublicKey, pub_key),
+        (TLV.kTLVType_SessionID, session_id),
+        (TLV.kTLVType_EncryptedData, auth_tag),
+    ]
+
+
+def resume_m3(
+    pub_key: bytes, derive: Callable[[bytes, bytes], bytes], response: dict[int, bytes]
+) -> Callable[[bytes, bytes], bytes] | None:
+    if not (method := response.get(TLV.kTLVType_Method)):
+        logger.debug("M3: Failure to resume existing session: Method not present")
+        return None
+
+    if int.from_bytes(method, "little") != TLV.kTLVMethod_Resume:
+        logger.debug("M3: Failure to resume existing session: Method != Resume")
+        return None
+
+    if not (session_id := response.get(TLV.kTLVType_SessionID)):
+        logger.debug("M3: Failure to resume existing session: No session id present")
+        return None
+
+    if not (auth_tag := response.get(TLV.kTLVType_EncryptedData)):
+        logger.debug("M3: Failure to resume existing session: No auth tag present")
+        return None
+
+    response_key = derive(
+        pub_key + session_id,
+        b"Pair-Resume-Response-Info",
+    )
+
+    key = ChaCha20Poly1305Decryptor(response_key)
+    plaintext = key.decrypt(
+        b"",
+        b"PR-Msg02",
+        bytes([0, 0, 0, 0]),
+        auth_tag,
+    )
+
+    if plaintext != b"":
+        logger.debug(
+            "M3: Failure to resume existing session: Could not decrypt kTLVType_EncryptedData"
+        )
+        return None
+
+    shared_secret = derive(
+        pub_key + session_id,
+        b"Pair-Resume-Shared-Secret-Info",
+    )
+
+    def derive(salt: bytes, info: bytes, length: int = 32) -> bytes:
+        return hkdf_derive(shared_secret, salt, info, length=length)
+
+    logger.debug("M3: Resume exchange success")
+
+    return session_id, derive
+
+
 def get_session_keys(
-    pairing_data: dict[str, str | int | list[Any]]
+    pairing_data: dict[str, str | int | list[Any]],
+    session_id=None,
+    derive=None,
 ) -> Generator[
     (
         tuple[list[tuple[int, bytearray] | tuple[int, bytes]], list[int]]
@@ -380,6 +461,10 @@ def get_session_keys(
 
     request_tlv = [(TLV.kTLVType_State, TLV.M1), (TLV.kTLVType_PublicKey, ios_key_pub)]
 
+    # If session_id is provided we can request that the accessory resumes it
+    if session_id and derive:
+        request_tlv = resume_m1(session_id, ios_key_pub, derive)
+
     step2_expectations = [
         TLV.kTLVType_State,
         TLV.kTLVType_PublicKey,
@@ -393,6 +478,10 @@ def get_session_keys(
     response_tlv = dict(response_tlv)
     handle_state_step(response_tlv, TLV.M2)
 
+    # Check for kTLVMethod_Resume - we might be able to shortcut the pair-verify
+    if derive and (response := resume_m3(ios_key_pub, derive, response_tlv)):
+        return response
+
     if TLV.kTLVType_PublicKey not in response_tlv:
         raise InvalidError("M2: Missing public key")
 
@@ -408,13 +497,13 @@ def get_session_keys(
 
     # 2) derive session key
     session_key = hkdf_derive(
-        shared_secret, "Pair-Verify-Encrypt-Salt", "Pair-Verify-Encrypt-Info"
+        shared_secret, b"Pair-Verify-Encrypt-Salt", b"Pair-Verify-Encrypt-Info"
     )
 
     # 3) verify auth tag on encrypted data and 4) decrypt
     encrypted = response_tlv[TLV.kTLVType_EncryptedData]
-    decrypted = chacha20_aead_decrypt(
-        bytes(), session_key, b"PV-Msg02", bytes([0, 0, 0, 0]), encrypted
+    decrypted = ChaCha20Poly1305Decryptor(session_key).decrypt(
+        b"", b"PV-Msg02", bytes([0, 0, 0, 0]), encrypted
     )
     if type(decrypted) == bool and not decrypted:
         raise InvalidAuthTagError("step 3")
@@ -475,8 +564,8 @@ def get_session_keys(
     )
 
     # 10) encrypt and sign
-    encrypted_data_with_auth_tag = chacha20_aead_encrypt(
-        bytes(), session_key, b"PV-Msg03", bytes([0, 0, 0, 0]), sub_tlv
+    encrypted_data_with_auth_tag = ChaCha20Poly1305Encryptor(session_key).encrypt(
+        b"", b"PV-Msg03", bytes([0, 0, 0, 0]), sub_tlv
     )
 
     # 11) create tlv
@@ -495,7 +584,13 @@ def get_session_keys(
     handle_state_step(response_tlv, TLV.M4)
 
     # return function to calculate session keys
-    def derive(salt, info):
-        return hkdf_derive(shared_secret, salt, info)
+    def derive(salt: bytes, info: bytes, length: int = 32) -> bytes:
+        return hkdf_derive(shared_secret, salt, info, length=length)
 
-    return derive
+    session_id = derive(
+        b"Pair-Verify-ResumeSessionID-Salt",
+        b"Pair-Verify-ResumeSessionID-Info",
+        length=8,
+    )
+
+    return session_id, derive
