@@ -18,14 +18,21 @@ from __future__ import annotations
 import base64
 from dataclasses import dataclass
 import logging
+from typing import AsyncIterable
 
 from aiohomekit import exceptions
-from aiohomekit.controller import Controller
-from aiohomekit.controller.ip.connection import HomeKitConnection
-from aiohomekit.controller.pairing import AbstractPairing
+from aiohomekit.characteristic_cache import CharacteristicCacheMemory
+from aiohomekit.controller.abstract import (
+    AbstractController,
+    AbstractDiscovery,
+    AbstractPairing,
+    FinishPairing,
+)
 from aiohomekit.exceptions import AccessoryNotFoundError
-from aiohomekit.model import Accessories
+from aiohomekit.model import Accessories, AccessoriesState
+from aiohomekit.model.categories import Categories
 from aiohomekit.model.characteristics import CharacteristicsTypes
+from aiohomekit.model.status_flags import StatusFlags
 from aiohomekit.protocol.statuscodes import HapStatusCode
 from aiohomekit.uuid import normalize_uuid
 
@@ -43,13 +50,13 @@ class FakeDescription:
     name: str = "TestDevice"
     id: str = "00:00:00:00:00:00"
     model: str = "TestDevice"
-    status_flags: int = 1
+    status_flags: StatusFlags = StatusFlags.UNPAIRED
     config_num: int = 1
     state_num: int = 1
-    category: str = "Other"
+    category: Categories = Categories.OTHER
 
 
-class FakeDiscovery:
+class FakeDiscovery(AbstractDiscovery):
 
     description = FakeDescription()
 
@@ -62,62 +69,38 @@ class FakeDiscovery:
         self.pairing_code = "111-22-333"
 
     @property
-    def info(self):
-        sf = 0
-
-        # Is accessory unpaired?
+    def status_flags(self) -> StatusFlags:
         if self.description.id not in self.controller.pairings:
-            sf = sf | 0x01
+            return StatusFlags.UNPAIRED
+        return StatusFlags(0)
 
-        return {
-            "name": "TestDevice",
-            "address": "127.0.0.1",
-            "port": 8080,
-            "md": "TestDevice",
-            "pv": "1.0",
-            "id": self.description.id,
-            "c#": 1,
-            "s#": 1,
-            "ff": 0,
-            "ci": 0,
-            "sf": sf,
-        }
-
-    @property
-    def paired(self) -> bool:
-        return self.description.id in self.controller.pairings
-
-    async def perform_pairing(self, alias: str, pin):
-        finish_pairing = await self.start_pairing(alias)
-        return await finish_pairing(pin)
-
-    async def async_start_pairing(self, alias: str):
-        return await self.start_pairing(alias)
-
-    async def start_pairing(self, alias: str):
+    async def async_start_pairing(self, alias: str) -> FinishPairing:
         if self.description.id in self.controller.pairings:
             raise exceptions.AlreadyPairedError(f"{self.description.id} already paired")
 
-        async def finish_pairing(pairing_code):
+        async def finish_pairing(pairing_code: str) -> FakePairing:
             if pairing_code != self.pairing_code:
                 raise exceptions.AuthenticationError("M4")
-            pairing_data = {}
-            pairing_data["AccessoryIP"] = self.info["address"]
-            pairing_data["AccessoryPort"] = self.info["port"]
-            pairing_data["Connection"] = "IP"
 
-            obj = self.controller.pairings[alias] = FakePairing(
-                self.controller, pairing_data, self.accessories
-            )
+            discovery = self.controller.discoveries[self.description.id]
+            discovery.description = FakeDescription(status_flags=0)
+
+            pairing_data = {}
+            # pairing_data["AccessoryIP"] = self.address
+            # pairing_data["AccessoryPort"] = self.port
+            pairing_data["AccessoryPairingID"] = discovery.description.id
+            pairing_data["Connection"] = "Fake"
+
+            obj = FakePairing(self.controller, pairing_data, self.accessories)
+            self.controller.aliases[alias] = obj
+            self.controller.pairings[discovery.description.id] = obj
+
             return obj
 
         return finish_pairing
 
-    async def identify(self):
-        return True
-
-    async def async_identify(self):
-        return True
+    async def async_identify(self) -> None:
+        """Trigger an identify routinue."""
 
 
 class PairingTester:
@@ -216,20 +199,10 @@ class FakePairing(AbstractPairing):
         """Create a fake pairing from an accessory model."""
         super().__init__(controller)
 
-        class MockConnection(HomeKitConnection):
-            @property
-            def is_connected(self) -> bool:
-                raise ValueError("This attribute is removed")
-
-            async def reconnect_soon(self) -> None:
-                pass
-
-        self.connection = MockConnection(None, "fake_host", 1234)
-        self.connection.transport = "mock_transport"
-        self.connection.protocol = "mock_protocol"
+        self.id = pairing_data["AccessoryPairingID"]
 
         self.accessories = accessories
-        self.pairing_data: dict[str, AbstractPairing] = {}
+        self.pairing_data: dict[str, str] = {}
         self.available = True
 
         self.testing = PairingTester(self)
@@ -239,16 +212,36 @@ class FakePairing(AbstractPairing):
         return True
 
     async def close(self):
-        pass
+        """Close the connection."""
 
     async def identify(self):
-        pass
+        """Identify the accessory."""
 
     async def list_pairings(self):
+        """List pairing."""
         return []
 
     async def remove_pairing(self, pairing_id):
-        pass
+        """Remove a pairing."""
+
+    async def async_populate_accessories_state(
+        self, force_update: bool = False
+    ) -> bool:
+        """Populate the state of all accessories.
+
+        This method should try not to fetch all the accessories unless
+        we know the config num is out of date or force_update is True
+        """
+        if not self._accessories or force_update:
+            await self.list_accessories_and_characteristics()
+        return True
+
+    async def _process_config_changed(self, config_num: int) -> None:
+        await self.list_accessories_and_characteristics()
+        self._accessories_state = AccessoriesState(
+            self._accessories_state.accessories, config_num
+        )
+        self._callback_and_save_config_changed(config_num)
 
     async def list_accessories_and_characteristics(self):
         """Fake implementation of list_accessories_and_characteristics."""
@@ -288,7 +281,7 @@ class FakePairing(AbstractPairing):
         return base64.b64decode(FAKE_CAMERA_IMAGE)
 
 
-class FakeController(Controller):
+class FakeController(AbstractController):
     """
     A test fake that pretends to be a paired HomeKit accessory.
 
@@ -296,10 +289,13 @@ class FakeController(Controller):
     class.
     """
 
+    started: bool
+    discoveries: dict[str, FakeDiscovery]
+    pairings: dict[str, FakePairing]
+    aliases: dict[str, FakePairing]
+
     def __init__(self):
-        """Create a Fake controller with no pairings."""
-        self.pairings = {}
-        self.discoveries = {}
+        super().__init__(char_cache=CharacteristicCacheMemory())
 
     def add_device(self, accessories):
         device_id = "00:00:00:00:00:00"
@@ -312,29 +308,34 @@ class FakeController(Controller):
 
     async def add_paired_device(self, accessories: Accessories, alias: str = None):
         discovery = self.add_device(accessories)
-        finish_pairing = await discovery.start_pairing(alias or discovery.device_id)
+        finish_pairing = await discovery.async_start_pairing(
+            alias or discovery.description.id
+        )
         return await finish_pairing(discovery.pairing_code)
 
-    async def async_discover(self, max_seconds: int = 10):
+    async def async_start(self) -> None:
+        self.started = True
+
+    async def async_stop(self) -> None:
+        self.started = False
+
+    async def async_discover(
+        self, max_seconds: int = 10
+    ) -> AsyncIterable[AbstractDiscovery]:
         for discovery in self.discoveries.values():
             yield discovery
 
-    async def async_find(self, device_id, max_seconds=10):
+    async def async_find(self, device_id, max_seconds=10) -> AbstractDiscovery:
         try:
             return self.discoveries[device_id]
         except KeyError:
             raise AccessoryNotFoundError(device_id)
 
-    async def discover_ip(self, max_seconds: int = 10):
-        return await self.discoveries.values()
-
-    async def find_ip_by_device_id(self, device_id, max_seconds=10):
-        return await self.async_find(device_id)
-
     async def remove_pairing(self, alias: str) -> None:
-        del self.pairings[alias]
+        del self.pairings[self.aliases[alias].id]
+        del self.aliases[alias]
 
     def load_pairing(self, alias: str, pairing_data):
         # This assumes a test has already preseed self.pairings with a fake via
         # add_paired_device
-        return self.pairings[alias]
+        return self.aliases[alias]
